@@ -1,16 +1,17 @@
-import abc
+from typing import Optional, List, TYPE_CHECKING
+from datetime import datetime
 from .chatgetter import ChatGetter
 from .sendergetter import SenderGetter
 from .messagebutton import MessageButton
 from .forward import Forward
 from .file import File
-from .. import TLObject, types, functions
+from .. import TLObject, types, functions, alltlobjects
 from ... import utils, errors
 
 
 # TODO Figure out a way to have the code generator error on missing fields
 # Maybe parsing the init function alone if that's possible.
-class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
+class Message(ChatGetter, SenderGetter, TLObject):
     """
     This custom class aggregates both :tl:`Message` and
     :tl:`MessageService` to ease accessing their members.
@@ -64,6 +65,12 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         pinned (`bool`):
             Whether this message is currently pinned or not.
 
+        noforwards (`bool`):
+            Whether this message can be forwarded or not.
+
+        invert_media (`bool`):
+            Whether the media in this message should be inverted.
+
         id (`int`):
             The ID of this message. This field is *always* present.
             Any other member is optional and may be `None`.
@@ -140,42 +147,76 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             (photo albums or video albums), all of them will
             have the same value here.
 
+        reactions (:tl:`MessageReactions`)
+            Reactions to this message.
+
         restriction_reason (List[:tl:`RestrictionReason`])
             An optional list of reasons why this message was restricted.
             If the list is `None`, this message has not been restricted.
 
+        ttl_period (`int`):
+            The Time To Live period configured for this message.
+            The message should be erased from wherever it's stored (memory, a
+            local database, etc.) when
+            ``datetime.now() > message.date + timedelta(seconds=message.ttl_period)``.
+
         action (:tl:`MessageAction`):
             The message action object of the message for :tl:`MessageService`
             instances, which will be `None` for other types of messages.
+
+        saved_peer_id (:tl:`Peer`)
     """
 
     # region Initialization
 
     def __init__(
             # Common to all
-            self, id,
+            self, id: int,
 
             # Common to Message and MessageService (mandatory)
-            peer_id=None, date=None,
+            peer_id: types.TypePeer = None,
+            date: Optional[datetime] = None,
 
             # Common to Message and MessageService (flags)
-            out=None, mentioned=None, media_unread=None, silent=None,
-            post=None, from_id=None, reply_to=None,
+            out: Optional[bool] = None,
+            mentioned: Optional[bool] = None,
+            media_unread: Optional[bool] = None,
+            silent: Optional[bool] = None,
+            post: Optional[bool] = None,
+            from_id: Optional[types.TypePeer] = None,
+            reply_to: Optional[types.TypeMessageReplyHeader] = None,
+            ttl_period: Optional[int] = None,
 
             # For Message (mandatory)
-            message=None,
+            message: Optional[str] = None,
 
             # For Message (flags)
-            fwd_from=None, via_bot_id=None, media=None, reply_markup=None,
-            entities=None, views=None, edit_date=None, post_author=None,
-            grouped_id=None, from_scheduled=None, legacy=None,
-            edit_hide=None, pinned=None, restriction_reason=None,
-            forwards=None, replies=None,
+            fwd_from: Optional[types.TypeMessageFwdHeader] = None,
+            via_bot_id: Optional[int] = None,
+            media: Optional[types.TypeMessageMedia] = None,
+            reply_markup: Optional[types.TypeReplyMarkup] = None,
+            entities: Optional[List[types.TypeMessageEntity]] = None,
+            views: Optional[int] = None,
+            edit_date: Optional[datetime] = None,
+            post_author: Optional[str] = None,
+            grouped_id: Optional[int] = None,
+            from_scheduled: Optional[bool] = None,
+            legacy: Optional[bool] = None,
+            edit_hide: Optional[bool] = None,
+            pinned: Optional[bool] = None,
+            noforwards: Optional[bool] = None,
+            invert_media: Optional[bool] = None,
+            reactions: Optional[types.TypeMessageReactions] = None,
+            restriction_reason: Optional[List[types.TypeRestrictionReason]] = None,
+            forwards: Optional[int] = None,
+            replies: Optional[types.TypeMessageReplies] = None,
+            saved_peer_id: Optional[types.TypePeer] = None,
 
             # For MessageAction (mandatory)
-            action=None):
+            action: Optional[types.TypeMessageAction] = None
+    ):
         # Common properties to messages, then to service (in the order they're defined in the `.tl`)
-        self.out = out
+        self.out = bool(out)
         self.mentioned = mentioned
         self.media_unread = media_unread
         self.silent = silent
@@ -199,9 +240,14 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         self.replies = replies
         self.edit_date = edit_date
         self.pinned = pinned
+        self.noforwards = noforwards
+        self.invert_media = invert_media
         self.post_author = post_author
         self.grouped_id = grouped_id
+        self.reactions = reactions
         self.restriction_reason = restriction_reason
+        self.ttl_period = ttl_period
+        self.saved_peer_id = saved_peer_id
         self.action = action
 
         # Convenient storage for custom functions
@@ -216,6 +262,7 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         self._via_bot = None
         self._via_input_bot = None
         self._action_entities = None
+        self._linked_chat = None
 
         sender_id = None
         if from_id is not None:
@@ -233,6 +280,8 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         SenderGetter.__init__(self, sender_id)
 
         self._forward = None
+        self._reply_to_chat = None
+        self._reply_to_sender = None
 
     def _finish_init(self, client, entities, input_chat):
         """
@@ -241,7 +290,13 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         known entities.
         """
         self._client = client
-        cache = client._entity_cache
+
+        # Make messages sent to ourselves outgoing unless they're forwarded.
+        # This makes it consistent with official client's appearance.
+        if self.peer_id == types.PeerUser(client._self_id) and not self.fwd_from:
+            self.out = True
+
+        cache = client._mb_entity_cache
 
         self._sender, self._input_sender = utils._get_entity_pair(
             self.sender_id, entities, cache)
@@ -276,6 +331,19 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
                 self._action_entities = [entities.get(utils.get_peer_id(
                     types.PeerChat(self.action.chat_id)))]
 
+        if self.replies and self.replies.channel_id:
+            self._linked_chat = entities.get(utils.get_peer_id(
+                    types.PeerChannel(self.replies.channel_id)))
+        
+        if isinstance(self.reply_to, types.MessageReplyHeader):
+            if self.reply_to.reply_to_peer_id:
+                self._reply_to_chat = entities.get(utils.get_peer_id(self.reply_to.reply_to_peer_id))
+            if self.reply_to.reply_from:
+                if self.reply_to.reply_from.from_id:
+                    self._reply_to_sender = entities.get(utils.get_peer_id(self.reply_to.reply_from.from_id))
+
+
+
     # endregion Initialization
 
     # region Public Properties
@@ -350,6 +418,22 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         """
         return self._forward
 
+    @property
+    def reply_to_chat(self):
+        """
+        The :tl:`Channel` in which the replied-to message was sent,
+        if this message is a reply in another chat
+        """
+        return self._reply_to_chat
+
+    @property
+    def reply_to_sender(self):
+        """
+        The :tl:`User`, :tl:`Channel`, or whatever other entity that
+        sent the replied-to message, if this message is a reply in another chat.
+        """
+        return self._reply_to_sender
+
     @property
     def buttons(self):
         """
@@ -735,12 +819,26 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
 
     async def edit(self, *args, **kwargs):
         """
-        Edits the message iff it's outgoing. Shorthand for
+        Edits the message if it's outgoing. Shorthand for
         `telethon.client.messages.MessageMethods.edit_message`
         with both ``entity`` and ``message`` already set.
 
-        Returns `None` if the message was incoming,
-        or the edited `Message` otherwise.
+        Returns
+            The edited `Message <telethon.tl.custom.message.Message>`,
+            unless `entity` was a :tl:`InputBotInlineMessageID` or :tl:`InputBotInlineMessageID64` in which
+            case this method returns a boolean.
+
+        Raises
+            ``MessageAuthorRequiredError`` if you're not the author of the
+            message but tried editing it anyway.
+
+            ``MessageNotModifiedError`` if the contents of the message were
+            not modified at all.
+
+            ``MessageIdInvalidError`` if the ID of the message is invalid
+            (the ID itself may be correct, but the message with that ID
+            cannot be edited). For example, when trying to edit messages
+            with a reply markup (or clear markup) this error will be raised.
 
         .. note::
 
@@ -754,9 +852,6 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             This is generally the most desired and convenient behaviour,
             and will work for link previews and message buttons.
         """
-        if self.fwd_from or not self.out or not self._client:
-            return None  # We assume self.out was patched for our chat
-
         if 'link_preview' not in kwargs:
             kwargs['link_preview'] = bool(self.web_preview)
 
@@ -799,7 +894,7 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
 
     async def click(self, i=None, j=None,
                     *, text=None, filter=None, data=None, share_phone=None,
-                    share_geo=None):
+                    share_geo=None, password=None):
         """
         Calls :tl:`SendVote` with the specified poll option
         or `button.click <telethon.tl.custom.messagebutton.MessageButton.click>`
@@ -878,6 +973,12 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
 
                 If the button is pressed without this, `ValueError` is raised.
 
+            password (`str`):
+                When clicking certain buttons (such as BotFather's confirmation
+                button to transfer ownership), if your account has 2FA enabled,
+                you need to provide your account's password. Otherwise,
+                `teltehon.errors.PasswordHashInvalidError` is raised.
+
             Example:
 
                 .. code-block:: python
@@ -901,19 +1002,13 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             return
 
         if data:
-            if not await self.get_input_chat():
+            chat = await self.get_input_chat()
+            if not chat:
                 return None
 
-            try:
-                return await self._client(
-                    functions.messages.GetBotCallbackAnswerRequest(
-                        peer=self._input_chat,
-                        msg_id=self.id,
-                        data=data
-                    )
-                )
-            except errors.BotResponseTimeoutError:
-                return None
+            but = types.KeyboardButtonCallback('', data)
+            return await MessageButton(self._client, but, chat, None, self.id).click(
+                share_phone=share_phone, share_geo=share_geo, password=password)
 
         if sum(int(x is not None) for x in (i, text, filter)) >= 2:
             raise ValueError('You can only set either of i, text or filter')
@@ -945,7 +1040,7 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
 
             options = find_options()
             if options is None:
-                return
+                options = []
             return await self._client(
                 functions.messages.SendVoteRequest(
                     peer=self._input_chat,
@@ -985,7 +1080,8 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
 
         button = find_button()
         if button:
-            return await button.click(share_phone=share_phone, share_geo=share_geo)
+            return await button.click(
+                share_phone=share_phone, share_geo=share_geo, password=password)
 
     async def mark_read(self):
         """
@@ -998,7 +1094,7 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             await self._client.send_read_acknowledge(
                 await self.get_input_chat(), max_id=self.id)
 
-    async def pin(self, *, notify=False):
+    async def pin(self, *, notify=False, pm_oneside=False):
         """
         Pins the message. Shorthand for
         `telethon.client.messages.MessageMethods.pin_message`
@@ -1009,7 +1105,7 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         #      That or figure out a way to always set it directly.
         if self._client:
             return await self._client.pin_message(
-                await self.get_input_chat(), self.id, notify=notify)
+                await self.get_input_chat(), self.id, notify=notify, pm_oneside=pm_oneside)
 
     async def unpin(self):
         """
@@ -1085,10 +1181,12 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
                         bot = self.input_sender
                         if not bot:
                             raise ValueError('No input sender')
+                        return bot
                     else:
                         try:
-                            return self._client._entity_cache[self.via_bot_id]
-                        except KeyError:
+                            return self._client._mb_entity_cache.get(
+                                utils.resolve_id(self.via_bot_id)[0])._as_input_peer()
+                        except AttributeError:
                             raise ValueError('No input sender') from None
 
     def _document_by_attribute(self, kind, condition=None):
